//******************************************************************************
// Sketch for mega2560 board to read the endless pot encoders, corresponding   *
// buttons, stop, play, rec, mute buttons ("Playback_switch") and their LEDs,  *
// navigation buttons, function buttons and rotary encoder.                    *
// Data sent over i2C.                                                         *
// v 1.0 initial version                                                       *
// v 1.1 Changed to store MIDI data values (0-127, 0/127) directly instead of  *
//       0-255 and 0-1 for pots and buttons respectively. Done at wire.write   *
//******************************************************************************

//******************************************************************************
//  Includes and dependencies                                                  *
//******************************************************************************
#include <Wire.h>                   // Library for I2C communication

//******************************************************************************
// Constant value definitions
//******************************************************************************
#define ADC_HYSTERESIS        15    // Must be 1 or higher. Noise filter, determines how big ADC change needed
#define MAX_POT_VALUE         255
#define POT_SENSITIVITY       7     // Higher number = more turns needed to reach max value
#define ADC_MAX_VALUE         1023
#define SWITCHDEBOUNCE        30    // the debounce time in ms, increase if the output flickers

//******************************************************************************
// pin mapping constants for everything except the buttons                     *
//******************************************************************************
// Constant declarations for Analog pin mappings
const byte POT_ANALOG_INPUTS[16]  = {54,55,56,57,58,59,60,61,62,63,64,65,66,67,68,69};
// Constant declarations for LEDs
const byte LED[12]   = {49,48,47,46,45,44,43,42,53,52,51,50}; //LED[1:8]= toggle buttons, LED[9:12] Playback control buttons

//******************************************************************************
// data type definitions
//******************************************************************************
typedef struct                  //define struct with all values related to each 360 pot
{
  int ValuePotA = 0;            //Pot1 tap A value
  int ValuePotB = 0;            //Pot1 tap B value
  int PreviousValuePotA = 0;    //Used to remember last value to determine turn direction
  int PreviousValuePotB = 0;    //Used to remember last value to determine turn direction
  int DirPotA = 1;              //Direction for Pot 1 tap A
  int DirPotB = 1;              //Direction for Pot1 tap B
  int Direction  = 1;           //Final CALCULATED direction
  int Value = 0;                //Final CALCULATED value
} pot_360;

typedef struct                  // define struct with all values related to each switch
{
  bool state;                   // Current state
  bool previousState;           // Previous state
  bool toggle;                  // Toggle function or momentary: TRUE = Toggle
  long debounceTimeStamp;       // Timestamp for debounce
  byte pin;                     // Input pin
} button;

//******************************************************************************
//  Global variable definitions                                                *
//******************************************************************************
pot_360 encoder[8];         // define an array holding the variables for all 8 pots

button switches[22] =       // define an array holding the variables and config for all 22 buttons {state,Prev state, Toggle, Time, pin}
{                           // All buttons with associated LED control must be first in the array
  {false,false,true,0,22},  // switches[0] = toggle switch 1 (with LED)
  {false,false,true,0,23},  // switches[1] = toggle switch 2 (with LED)
  {false,false,true,0,24},  // switches[2] = toggle switch 3 (with LED)
  {false,false,true,0,25},  // switches[3] = toggle switch 4 (with LED)
  {false,false,true,0,26},  // switches[4] = toggle switch 5 (with LED)
  {false,false,true,0,27},  // switches[5] = toggle switch 6 (with LED)
  {false,false,true,0,28},  // switches[6] = toggle switch 7 (with LED)
  {false,false,true,0,29},  // switches[7] = toggle switch 8 (with LED)
  {false,false,false,0,10}, // switches[8] = Playback switch 1 STOP (with LED)
  {false,false,false,0,11}, // switches[9] = Playback switch 2 PLAY (with LED)
  {false,false,true,0,12},  // switches[10] = Playback switch 3 REC (with LED)
  {false,false,false,0,13}, // switches[11] = Playback switch 4 SPEAKER (with LED)
  {false,false,false,0,6},  // switches[12] = Navig switch 1 LEFT
  {false,false,false,0,7},  // switches[13] = Navig switch 2 UP
  {false,false,false,0,8},  // switches[14] = Navig switch 3 DOWN
  {false,false,false,0,9},  // switches[15] = Navig switch 4 RIGHT
  {false,false,false,0,35}, // switches[19] = Function switch Control
  {false,false,false,0,32}, // switches[16] = Function switch Sound
  {false,false,false,0,36}, // switches[20] = Function switch F1
  {false,false,false,0,33}, // switches[17] = Function switch F2
  {false,false,false,0,37}, // switches[21] = Function switch Enter
  {false,false,false,0,34}  // switches[18] = Function switch ESC
};

//******************************************************************************
//  Function declarations                                                      *
//******************************************************************************
bool decode360pot(pot_360* potmeter, char inputA, char inputB);  //define function for 360 pot decoding

//******************************************************************************
// Setup code                                                                  *
//******************************************************************************
void setup()                    // Initialization code
{
  Serial.begin(115200);         // Set up debug uart baudrate
  Wire.begin();                 // join i2c bus (address optional for master)

  for (int i=0; i<8; i++)       // Loop for reading current potmeter readings
  {
    encoder[i].PreviousValuePotA = analogRead(POT_ANALOG_INPUTS[i*2]);   // Initialize the "previous" values by ADC sampling
    encoder[i].PreviousValuePotB = analogRead(POT_ANALOG_INPUTS[i*2+1]); // Initialize the "previous" values by ADC sampling
  }
  for (int i=0; i<12; i++)       // Loop for initializing all LED pins
  {
    digitalWrite(LED[i], HIGH);  // sets the LED to off
    pinMode(LED[i], OUTPUT);     // sets the LED ports to output
  }
}

//******************************************************************************
// Main loop                                                                   *
//******************************************************************************
void loop()
{
  //****************************************************************************
  // First read all pots and buttons                                           *
  //****************************************************************************
  // Decode all eight 360 pots
  bool inputChanged = false;
  for (int i=0; i<8; i++)
  {
    inputChanged |= decode360pot(&encoder[i], POT_ANALOG_INPUTS[i*2], POT_ANALOG_INPUTS[(i*2)+1] ); //decode pot and indicate if any pot value has changed
  }
  // read all the buttons, active high.
  for (int i=0; i<22; i++)
  {
    byte currentstate = digitalRead (switches[i].pin);  // Read current state of the pin

    if (switches[i].toggle == true)                     // if switch is toggle type
    {
      if ((currentstate == true) && (switches[i].previousState == false) && ((millis() - switches[i].debounceTimeStamp) > SWITCHDEBOUNCE)) //If switch is active, last time was inactive, and more than "SWITCHDEBOUNCE" ms since last toggle
      {
        if (switches[i].state == true)                  // If previous state was true, change to false
        {
          switches[i].state = false;
        }
        else
        {
          switches[i].state = true;
        }
        inputChanged |= true;                           // Indicate button toggled
        switches[i].debounceTimeStamp = millis();       // Reset timestamp for when toggle happened (for debouncing logic)
        if (i<sizeof(LED))                              // If its one of the  buttons who have associated LEDs (MUST be placed first in array)
        {
          digitalWrite(LED[i], !switches[i].state);
        }
//        switches[i].previousState = currentstate;       // Remember current state until next check
      }
    }
    else  //if switch is not toggle type, it should be momentary
    {
      if ((currentstate != switches[i].previousState) && ((millis() - switches[i].debounceTimeStamp) > SWITCHDEBOUNCE)) //If button has changed state and more than "SWITCHDEBOUNCE" ms since last press
      {
        switches[i].state = currentstate;               // Set state to current state
        inputChanged |= true;                           // Indicate button toggled
        switches[i].debounceTimeStamp = millis();       // Reset timestamp for when toggle happened (for debouncing logic)
        if (i<sizeof(LED))                              // If its one of the  buttons who have associated LEDs (MUST be placed first in array)
        {
          digitalWrite(LED[i], !switches[i].state);
        }
//        switches[i].previousState = currentstate;       // Remember current state until next check
      }
    }
    switches[i].previousState = currentstate;       // Remember current state until next check
  }

  //****************************************************************************
  // Second, check for any changes, and if so send data over I2C and debug port*                                           *
  //****************************************************************************
  Wire.beginTransmission(8); // transmit to device #8
  if (inputChanged)                     // If any pot or button has changed value
  {
    for (int i=0; i<8; i++)
    {
      Wire.write(encoder[i].Value/2);   // Send pot value over I2C. Divide by 2 to get MIDI values 0-127
      //Serial.print(encoder[i].Value);   // Send pot value to serial monitor
      //Serial.print(" , ");
    }

    for (int i=0; i<22; i++)
    {
      Wire.write(switches[i].state*127); // Send Toggle Switch value over I2C. Multiply with 127 to get 0/127 for false/true respectively
      //Serial.print(switches[i].state);   // Send Toggle Switch value to serial monitor
      //Serial.print(" , ");
    }
    Wire.endTransmission();              // I2C stop
    Serial.println();                    //New line
  }
}

//******************************************************************************
// Function/Algorithm, to decode 360 pot absed on the two individual pot taps  *
// The function require a pointer to the data for the potmeter, which is stored*
// in structs of type pot_360. It also require the ADC input pins. The function*
// returns a boolean which is true if any of the pots have changed value since    *
//******************************************************************************
bool decode360pot(pot_360* potmeter, char inputA, char inputB)
{
  /****************************************************************************
  * Step 1 decode each  individual pot tap's direction
  ****************************************************************************/
  potmeter->ValuePotA  = analogRead(inputA);
  potmeter->ValuePotB  = analogRead(inputB);
  // First check direction for Tap A
  if (potmeter->ValuePotA > (potmeter->PreviousValuePotA + ADC_HYSTERESIS))       // check if new reading is higher (by <debounce value>), if so...
  {
    potmeter->DirPotA = 1;                                              // ...direction of tap A is up
  }
  else if (potmeter->ValuePotA < (potmeter->PreviousValuePotA - ADC_HYSTERESIS))  // check if new reading is lower (by <debounce value>), if so...
  {
    potmeter->DirPotA = -1;                                             // ...direction of tap A is down
  }
  else
  {
    potmeter->DirPotA = 0;                                              // No change
  }
  // then check direction for tap B
  if (potmeter->ValuePotB > (potmeter->PreviousValuePotB + ADC_HYSTERESIS))       // check if new reading is higher (by <debounce value>), if so...
  {
    potmeter->DirPotB = 1;                                              // ...direction of tap B is up
  }
  else if (potmeter->ValuePotB < (potmeter->PreviousValuePotB - ADC_HYSTERESIS))  // check if new reading is lower (by <debounce value>), if so...
  {
    potmeter->DirPotB = -1;                                             // ...direction of tap B is down
  }
  else
  {
    potmeter->DirPotB = 0;                                              // No change
  }

  /****************************************************************************
  * Step 2: Determine actual direction of ENCODER based on each individual
  * potentiometer tap´s direction and the phase
  ****************************************************************************/
  if (potmeter->DirPotA == -1 and potmeter->DirPotB == -1)       //If direction of both taps is down
  {
    if (potmeter->ValuePotA > potmeter->ValuePotB)               // If value A above value B...
    {
      potmeter->Direction = 1;                         // ...direction is up
    }
    else
    {
      potmeter->Direction = -1;                        // otherwise direction is down
    }
  }
  else if (potmeter->DirPotA == 1 and potmeter->DirPotB == 1)    //If direction of both taps is up
  {
    if (potmeter->ValuePotA < potmeter->ValuePotB)               // If value A below value B...
    {
      potmeter->Direction = 1;                         // ...direction is up
    }
    else
    {
      potmeter->Direction = -1;                        // otherwise direction is down
    }
  }
  else if (potmeter->DirPotA == 1 and potmeter->DirPotB == -1)   // If A is up and B is down
  {
    if ( (potmeter->ValuePotA > (ADC_MAX_VALUE/2)) || (potmeter->ValuePotB > (ADC_MAX_VALUE/2)) )  //If either pot at upper range A/B = up/down means direction is up
    {
      potmeter->Direction = 1;
    }
    else                                     //otherwise if both pots at lower range A/B = up/down means direction is down
    {
      potmeter->Direction = -1;
    }
  }
  else if (potmeter->DirPotA == -1 and potmeter->DirPotB == 1)
  {
    if ( (potmeter->ValuePotA < (ADC_MAX_VALUE/2)) || (potmeter->ValuePotB < (ADC_MAX_VALUE/2)))   //If either pot  at lower range, A/B = down/up means direction is down
    {
      potmeter->Direction = 1;
    }
    else                                     //otherwise if bnoth pots at higher range A/B = up/down means direction is down
    {
      potmeter->Direction = -1;
    }
  }
  else
  {
    potmeter->Direction = 0;                 // if any of tap A or B has status unchanged (0), indicate unchanged
  }

  /****************************************************************************
  * Step 3: Calculate value based on direction, how big change in ADC value,
  * and sensitivity. Avoid values around zero and max  as value has flat region
  ****************************************************************************/
  if (potmeter->DirPotA != 0 && potmeter->DirPotB != 0)         //If both taps indicate movement
  {
    if ((potmeter->ValuePotA < ADC_MAX_VALUE*0.7) && (potmeter->ValuePotA > ADC_MAX_VALUE*0.3))         // if tap A is not at endpoints
    {
      potmeter->Value = potmeter->Value + potmeter->Direction*abs(potmeter->ValuePotA - potmeter->PreviousValuePotA)/POT_SENSITIVITY; //increment value
    }
    else                                    // If tap A is close to end points, use tap B to calculate value
    {
      potmeter->Value = potmeter->Value + potmeter->Direction*abs(potmeter->ValuePotB - potmeter->PreviousValuePotB)/POT_SENSITIVITY;  //Make sure to add/subtract at least 1, and then additionally the jump in voltage
    }
    // Finally apply output value limit control
    if (potmeter->Value < 0)               // If value decremented below zero, keep at zero
    {
      potmeter->Value = 0;
    }
    else if (potmeter->Value > MAX_POT_VALUE)  // If value increased above MAX_POT_VALUE, stop at MAX_POT_VALUE
    {
      potmeter->Value = MAX_POT_VALUE;
    }
    potmeter->PreviousValuePotA = potmeter->ValuePotA;          // Update previous value variable
    potmeter->PreviousValuePotB = potmeter->ValuePotB;          // Update previous value variable
    return true;                          // Indicate that there has been a value change by returning "true"
  }
  else
  {
    return false;                         // Indicate that there has not been any value change
  }
}
